"""Recording classes"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import logging

import bluepyopt as nrp
from . import responses

logger = logging.getLogger(__name__)


class Recording(object):

    """Response to stimulus"""

    def __init__(self, name=None, value=None, frozen=None):
        """Constructor"""

        self.name = name
        self.value = value
        self.frozen = frozen


class CompRecording(Recording):

    """Response to stimulus"""

    def __init__(
            self,
            name=None,
            value=None,
            frozen=None,
            location=None,
            variable='v'):
        """Constructor"""

        Recording.__init__(self, name=name, value=value, frozen=frozen)
        self.location = location
        self.variable = variable

        self.varvector = None
        self.tvector = None

        self.time = None
        self.voltage = None

        self.instantiated = False

    @property
    def response(self):
        """Return recording response"""

        if not self.instantiated:
            raise Exception(
                'Protocol not instantiated before requesting response')

        return responses.TimeVoltageResponse(self.name,
                                             self.tvector.to_python(),
                                             self.varvector.to_python())

    def instantiate(self, cell):
        """Instantiate recording"""

        logger.debug('Adding compartment recording of %s at %s',
                     self.variable, self.location)

        self.varvector = nrp.neuron.h.Vector()
        seg = self.location.instantiate(cell)
        self.varvector.record(getattr(seg, '_ref_%s' % self.variable))

        self.tvector = nrp.neuron.h.Vector()
        self.tvector.record(nrp.neuron.h._ref_t)  # pylint: disable=W0212

        self.instantiated = True

    def destroy(self):
        """Destroy recording"""

        self.varvector = None
        self.instantiated = False

    def __str__(self):
        """String representation"""

        return '%s: %s at %s' % (self.name, self.variable, self.location)



"""Morphology classes"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import os
import logging
logger = logging.getLogger(__name__)

import bluepyopt as nrp

# TODO define an addressing scheme


class Morphology(object):

    """Morphology class"""

    def __init__(self):
        """Constructor"""

        pass


class NrnFileMorphology(Morphology):

    """Morphology loaded from a file"""

    def __init__(self, morphology_path, do_replace_axon=False):
        """Constructor"""

        Morphology.__init__(self)
        # TODO speed up loading of morphologies from files
        # Path to morphology
        self.morphology_path = morphology_path
        self.do_replace_axon = do_replace_axon

    def __str__(self):
        """Return string representation"""

        return self.morphology_path

    def instantiate(self, cell):
        """Load morphology"""

        logger.debug('Loading morphology %s', self.morphology_path)

        if not os.path.exists(self.morphology_path):
            raise Exception(
                'Morphology not found at \'%s\'' %
                self.morphology_path)

        nrp.neuron.h.load_file('import3d.hoc')

        extension = self.morphology_path.split('.')[-1]

        if extension.lower() == 'swc':
            imorphology = nrp.neuron.h.Import3d_SWC_read()
        elif extension.lower() == 'asc':
            imorphology = nrp.neuron.h.Import3d_Neurolucida3()
        else:
            raise Exception("Unknown filetype: %s" % extension)

        # TODO this is to get rid of stdout print of neuron
        # probably should be more intelligent here, and filter out the
        # lines we don't want
        nrp.neuron.h.hoc_stdout('/dev/null')
        imorphology.input(self.morphology_path)
        nrp.neuron.h.hoc_stdout()

        morphology_importer = nrp.neuron.h.Import3d_GUI(imorphology, 0)

        morphology_importer.instantiate(cell.icell)

        # TODO replace these two functions with general function users can
        # specify
        if self.do_replace_axon:
            NrnFileMorphology.replace_axon(cell.icell)

        # TODO Set nseg should be called after all the parameters have been
        # set
        # (in case e.g. Ra was changed)
        NrnFileMorphology.set_nseg(cell.icell)

    def destroy(self):
        """Destroy morphology instantiation"""
        pass

    @staticmethod
    def set_nseg(icell):
        """Set the nseg of every section"""

        for section in icell.all:
            section.nseg = 1 + 2 * int(section.L / 40)

    @staticmethod
    def replace_axon(icell):
        """Replace axon"""

        ais_diams = [icell.axon[0].diam, icell.axon[0].diam]

        # Define origin of distance function
        nrp.neuron.h.distance(sec=icell.soma[0])
        for section in icell.axonal:
            # If distance to soma is larger than 60, store diameter
            if nrp.neuron.h.distance(0.5, sec=section) > 60:
                ais_diams[1] = section.diam
                break

        for section in icell.axonal:
            nrp.neuron.h.delete_section(sec=section)

        # Create new axon array
        nrp.neuron.h.execute('create axon[2]', icell)

        for index, section in enumerate(icell.axon):
            section.L = 30
            section.diam = ais_diams[index]
            icell.axonal.append(section)
            icell.all.append(section)

        icell.axon[0].connect(icell.soma[0], 1.0, 0.0)
        icell.axon[1].connect(icell.axon[0], 1.0, 0.0)

        logger.debug('Replace axon with AIS')



"""Run simple cell optimisation"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
# pylint: disable=R0914

import l5pc_template

import bluepyopt.electrical as nrpel
import bluepyopt.electrical.cellevaluator as ce


# TODO store definition dicts in json
# TODO rename 'score' into 'objective'
# TODO add functionality to read settings of every object from config format


def define_protocols():
    """Define protocols"""

    import json
    with open('protocols.json', 'r') as protocol_file:
        protocol_definitions = json.load(protocol_file)

    protocols = {}

    soma_loc = nrpel.locations.NrnSeclistCompLocation(
        name='soma',
        seclist_name='somatic',
        sec_index=0,
        comp_x=0.5)

    for protocol_name, protocol_definition in protocol_definitions.iteritems():
        # By default include somatic recording
        somav_recording = nrpel.recordings.CompRecording(
            name='%s.soma.v' %
            protocol_name,
            location=soma_loc,
            variable='v')

        recordings = [somav_recording]

        if 'extra_recordings' in protocol_definition:
            for recording_definition in protocol_definition['extra_recordings']:
                if recording_definition['type'] == 'somadistance':
                    location = nrpel.locations.NrnSomaDistanceCompLocation(
                        name=recording_definition['name'],
                        soma_distance=recording_definition['somadistance'],
                        seclist_name=recording_definition['seclist_name'])
                    var = recording_definition['var']
                    recording = nrpel.recordings.CompRecording(
                        name='%s.%s.%s' % (protocol_name, location.name, var),
                        location=location,
                        variable=recording_definition['var'])

                    recordings.append(recording)
                else:
                    raise Exception(
                        'Recording type %s not supported' %
                        recording_definition['type'])

        stimuli = []
        for stimulus_definition in protocol_definition['stimuli']:
            stimuli.append(nrpel.stimuli.NrnSquarePulse(
                step_amplitude=stimulus_definition['amp'],
                step_delay=stimulus_definition['delay'],
                step_duration=stimulus_definition['duration'],
                location=soma_loc,
                total_duration=stimulus_definition['totduration']))

        protocols[protocol_name] = nrpel.protocols.Protocol(
            protocol_name,
            stimuli,
            recordings)

    return protocols


def define_fitness_calculator(protocols):
    """Define fitness calculator"""

    import json
    with open('features.json', 'r') as protocol_file:
        feature_definitions = json.load(protocol_file)

    # TODO: add bAP stimulus
    objectives = []

    for protocol_name, locations in feature_definitions.iteritems():
        for location, features in locations.iteritems():
            for efel_feature_name, meanstd in features.iteritems():
                feature_name = '%s.%s.%s' % (
                    protocol_name, location, efel_feature_name)
                recording_names = {'': '%s.%s.v' % (protocol_name, location)}
                stimulus = protocols[protocol_name].stimuli[0]

                if location == 'soma':
                    threshold = -20
                    stim_start = stimulus.step_delay
                    stim_end = stimulus.step_delay + stimulus.step_duration
                elif 'dend' in location:
                    threshold = -30
                    stim_start = stimulus.step_delay
                    # bAP response can be after stimulus
                    stim_end = stimulus.total_duration

                feature = nrpel.efeatures.eFELFeature(
                    feature_name,
                    efel_feature_name=efel_feature_name,
                    recording_names=recording_names,
                    stim_start=stim_start,
                    stim_end=stim_end,
                    exp_mean=meanstd[0],
                    exp_std=meanstd[1],
                    threshold=threshold)
                objective = nrpel.objectives.SingletonObjective(
                    feature_name,
                    feature)
                objectives.append(objective)

    fitcalc = nrpel.scorecalculators.ObjectivesScoreCalculator(objectives)

    return fitcalc


def create():
    """Setup"""

    l5pc_cell = l5pc_template.create()

    fitness_protocols = define_protocols()
    fitness_calculator = define_fitness_calculator(fitness_protocols)

    param_names = [param.name
                   for param in l5pc_cell.params.values()
                   if not param.frozen]

    return ce.CellEvaluator(
        cell_template=l5pc_cell,
        param_names=param_names,
        fitness_protocols=fitness_protocols,
        fitness_calculator=fitness_calculator)



"""Location classes"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import bluepyopt as nrp


class Location(object):

    """EPhys feature"""

    def __init__(self, name):
        """Constructor"""

        self.name = name

# TODO make all these locations accept a cell name
# TODO instantiate should get the entire simulation environment
# TODO find better/more general name for this
# TODO specify in document abrevation comp=compartment, sec=section, ...


class NrnSeclistCompLocation(Location):

    """Compartment in a sectionlist"""

    def __init__(
            self,
            name,
            seclist_name=None,
            sec_index=None,
            comp_x=None):
        """Constructor"""

        Location.__init__(self, name)
        self.seclist_name = seclist_name
        self.sec_index = sec_index
        self.comp_x = comp_x

    def instantiate(self, cell):
        """Find the instantiate compartment"""

        import itertools

        isectionlist = getattr(cell.icell, self.seclist_name)

        # Get nth element of isectionlist
        # Sectionlists don't support direct indexing
        isection = next(
            itertools.islice(
                isectionlist,
                self.sec_index,
                self.sec_index + 1))
        icomp = isection(self.comp_x)

        return icomp

    def __str__(self):
        """String representation"""

        return '%s[%s](%s)' % (self.seclist_name, self.sec_index, self.comp_x)


class NrnSeclistLocation(Location):

    """Section in a sectionlist"""

    def __init__(
            self,
            name,
            seclist_name=None,
            sec_index=None):
        """Constructor"""

        Location.__init__(self, name)
        self.seclist_name = seclist_name

    def instantiate(self, cell):
        """Find the instantiate compartment"""

        isectionlist = getattr(cell.icell, self.seclist_name)

        return (isection for isection in isectionlist)

    def __str__(self):
        """String representation"""

        return '%s' % (self.seclist_name)


class NrnSeclistSecLocation(Location):

    """Section in a sectionlist"""

    def __init__(
            self,
            name,
            seclist_name=None,
            sec_index=None):
        """Constructor"""

        Location.__init__(self, name)
        self.seclist_name = seclist_name
        self.sec_index = sec_index

    def instantiate(self, cell):
        """Find the instantiate compartment"""

        import itertools

        isectionlist = getattr(cell.icell, self.secist_name)

        # Get nth element of isectionlist
        # Sectionlists don't support direct indexing
        isection = next(
            itertools.islice(
                isectionlist,
                self.sec_index,
                self.sec_index + 1))

        return isection

    def __str__(self):
        """String representation"""

        return '%s[%s]' % (self.seclist_name, self.sec_index)


class NrnSomaDistanceCompLocation(Location):

    """Compartment at distance from soma"""

    def __init__(self, name, soma_distance=None, seclist_name=None):
        """Constructor"""

        Location.__init__(self, name)
        self.soma_distance = soma_distance
        self.seclist_name = seclist_name

    # TODO this definitely has to be unit-tested
    # TODO add ability to specify origin
    # TODO rename 'seg' in 'compartment' everywhere
    def instantiate(self, cell):
        """Find the instantiate compartment"""

        soma = cell.icell.soma[0]

        nrp.neuron.h.distance(0, 0.5, sec=soma)

        iseclist = getattr(cell.icell, self.seclist_name)

        icomp = None
        max_diam = 0.0

        for isec in iseclist:
            start_distance = nrp.neuron.h.distance(1, 0.0, sec=isec)
            end_distance = nrp.neuron.h.distance(1, 1.0, sec=isec)

            min_distance = min(start_distance, end_distance)
            max_distance = max(start_distance, end_distance)

            if min_distance <= self.soma_distance <= end_distance:
                comp_x = float(self.soma_distance - min_distance) / \
                    (max_distance - min_distance)

                comp_diam = isec(comp_x).diam

                if comp_diam > max_diam:
                    icomp = isec(comp_x)

        if icomp is None:
            raise Exception(
                'No comp found at %s distance from soma' %
                self.soma_distance)

        return icomp

    def __str__(self):
        """String representation"""

        return '%f mum from soma in %s' % (
            self.soma_distance, self.seclist_name)



"""Run simple cell optimisation"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# pylint: disable=R0914

import pickle


def analyse_cp(opt=None, cp_filename=None):
    """Analyse optimisation results"""

    cp = pickle.load(open(cp_filename, "r"))
    results = (
        cp['population'],
        cp['halloffame'],
        cp['history'],
        cp['logbook'])

    _, hof, _, log = results

    print '\nHall of fame'
    print '################'
    for ind_number, individual in enumerate(hof[:1], 1):
        objectives = opt.evaluator.objective_dict(
            individual.fitness.values)
        parameter_values = opt.evaluator.param_dict(individual)

        print '\nIndividual %d' % ind_number
        print '#############'
        print 'Parameters: %s' % parameter_values
        print '\nObjective values: %s' % objectives

        fitness_protocols = opt.evaluator.fitness_protocols
        responses = opt.evaluator.\
            cell_template.run_protocols(fitness_protocols,
                                        parameter_values=parameter_values)

        plot_responses(responses)
        plot_objectives(objectives)

    plot_log(log)

    # plot_history(history)


def plot_log(log):
    """Plot logbook"""

    import matplotlib.pyplot as plt
    import numpy

    gen_numbers = log.select('gen')
    mean = log.select('avg')
    std = log.select('std')
    minimum = log.select('min')
    maximum = log.select('max')

    _, ax = plt.subplots(facecolor='white')

    ax.errorbar(
        gen_numbers,
        numpy.negative(mean),
        std,
        color='black',
        linewidth=2,
        label='mean/std')
    ax.plot(
        gen_numbers,
        numpy.negative(minimum),
        color='blue',
        linewidth=1,
        label='max')
    ax.plot(
        gen_numbers,
        numpy.negative(maximum),
        color='red',
        linewidth=1,
        label='min')
    ax.set_xlim(min(gen_numbers) - 1, max(gen_numbers) + 1)
    ax.set_xlabel('Gen #')
    ax.set_ylabel('Fitness')
    ax.legend()


def plot_history(history):
    """Plot the history of the individuals"""

    import networkx

    import matplotlib.pyplot as plt
    plt.figure()

    graph = networkx.DiGraph(history.genealogy_tree)
    graph = graph.reverse()     # Make the grah top-down
    # colors = [\
    #        toolbox.evaluate(history.genealogy_history[i])[0] for i in graph]
    positions = networkx.graphviz_layout(graph, prog="dot")
    networkx.draw(graph, positions)


def plot_objectives(objectives):
    """Plot responses of the cell template"""

    import matplotlib.pyplot as plt

    _, ax = plt.subplots(facecolor='white')

    ax.barh(range(len(objectives.values())), objectives.values(), color='b')
    ax.set_yticks(
        [x + 0.5 for x in range(len(objectives.keys()))])
    ax.set_yticklabels(objectives.keys(), size='small')
    ax.set_ylim(0, len(objectives.values()))
    ax.set_xlabel('Objective value (# std)')
    ax.set_ylabel('Objectives')

    plt.tight_layout()
    plt.subplots_adjust(top=0.95)


def plot_responses(responses):
    """Plot responses of the cell template"""

    import matplotlib.pyplot as plt
    import numpy

    n_of_cols = 1
    n_of_rows = len(responses)
    _, axes_obj = plt.subplots(n_of_rows, n_of_cols, facecolor='white')
    axes = numpy.ravel(axes_obj)

    for rec_number, (_, recording) in \
            enumerate(responses.items()):
        # axes[rec_number].set_title(recording_name)
        recording.plot(axes[rec_number])
        axes[rec_number].legend()

        axes[rec_number].set_xlabel('Time (ms)')
        axes[rec_number].set_ylabel('Vm (mV)')

    plt.tight_layout()
    plt.subplots_adjust(top=0.95)


def analyse_releasecircuit_model(opt):
    """Analyse L5PC model from release circuit"""

    # Parameters in release circuit model
    parameters = {
        'gIhbar_Ih.basal': 0.000080,
        'gNaTs2_tbar_NaTs2_t.apical': 0.026145,
        'gSKv3_1bar_SKv3_1.apical': 0.004226,
        'gIhbar_Ih.apical': 0.000080,
        'gImbar_Im.apical': 0.000143,
        'gNaTa_tbar_NaTa_t.axonal': 3.137968,
        'gK_Tstbar_K_Tst.axonal': 0.089259,
        'gamma_CaDynamics_E2.axonal': 0.002910,
        'gNap_Et2bar_Nap_Et2.axonal': 0.006827,
        'gSK_E2bar_SK_E2.axonal': 0.007104,
        'gCa_HVAbar_Ca_HVA.axonal': 0.000990,
        'gK_Pstbar_K_Pst.axonal': 0.973538,
        'gSKv3_1bar_SKv3_1.axonal': 1.021945,
        'decay_CaDynamics_E2.axonal': 287.198731,
        'gCa_LVAstbar_Ca_LVAst.axonal': 0.008752,
        'gamma_CaDynamics_E2.somatic': 0.000609,
        'gSKv3_1bar_SKv3_1.somatic': 0.303472,
        'gSK_E2bar_SK_E2.somatic': 0.008407,
        'gCa_HVAbar_Ca_HVA.somatic': 0.000994,
        'gNaTs2_tbar_NaTs2_t.somatic': 0.983955,
        'gIhbar_Ih.somatic': 0.000080,
        'decay_CaDynamics_E2.somatic': 210.485284,
        'gCa_LVAstbar_Ca_LVAst.somatic': 0.000333
    }
    fitness_protocols = opt.evaluator.fitness_protocols

    responses = opt.evaluator.cell_template.run_protocols(
        fitness_protocols,
        parameter_values=parameters)

    objectives = opt.evaluator.objective_dict(
        opt.
        evaluator.
        fitness_calculator.calculate_scores(
            responses))

    plot_responses(responses)
    plot_objectives(objectives)



"""Parameter scaler classes"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import bluepyopt as nrp


class ParameterScaler(object):

    """Parameter scalers"""

    def __init__(self, name):
        """Constructor"""

        self.name = name

# TODO get rid of the 'segment' here


class NrnSegmentLinearScaler(ParameterScaler):

    """eFEL feature"""

    def __init__(
            self,
            name=None,
            multiplier=1.0,
            offset=0.0):
        """Constructor"""

        ParameterScaler.__init__(self, name)
        self.multiplier = multiplier
        self.offset = offset

    def scale(self, value, _):
        """Scale a value based on a segment"""

        return self.multiplier * value + self.offset


class NrnSegmentSomaDistanceScaler(ParameterScaler):

    """eFEL feature"""

    def __init__(
            self,
            name=None,
            distribution=None):
        """Constructor"""

        ParameterScaler.__init__(self, name)
        self.distribution = distribution

    def scale(self, value, segment):
        """Scale a value based on a segment"""

        # TODO soma needs other addressing scheme

        soma = segment.sec.cell().soma[0]

        # Initialise origin
        nrp.neuron.h.distance(0, 0.5, sec=soma)

        distance = nrp.neuron.h.distance(1, segment.x, sec=segment.sec)

        # Find something to generalise this
        import math  # pylint:disable=W0611 #NOQA

        # This eval is unsafe (but is it ever dangerous ?)
        # pylint: disable=W0123

        return eval(self.distribution.format(distance=distance, value=value))



"""Objective classes"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


class Objective(object):

    """EPhys feature"""

    def __init__(self, name, features=None):
        """Constructor"""

        self.name = name
        self.features = features

    def calculate_feature_scores(self, responses):
        """Calculate the scores for the individual features"""

        scores = []
        for feature in self.features:
            scores.append(feature.calculate_score(responses))

        return scores


class SingletonObjective(Objective):

    """Single eFEL feature"""

    def __init__(self, name, feature):
        """Constructor"""

        Objective.__init__(self, name, [feature])

    def calculate_score(self, responses):
        """Objective score"""

        return self.calculate_feature_scores(responses)[0]


class MaxObjective(Objective):

    """Max of list of eFEL feature"""

    def __init__(self, name, features):
        """Constructor"""

        Objective.__init__(self, name, features)

    @property
    def calculate_score(self, responses):
        """Objective score"""

        return max(self.calculate_feature_scores(responses))


class WeightedSumObjective(Objective):

    """Weighted sum of list of eFEL features"""

    def __init__(self, name, features, weights):
        """Constructor"""

        Objective.__init__(self, name, features)
        if len(weights) != len(features):
            raise Exception(
                'WeightedSumObjective: number of weights must be equal to '
                'number of features')
        self.weights = weights

    @property
    def calculate_score(self, responses):
        """Objective score"""

        score = 0.0
        feature_scores = self.calculate_feature_scores(responses)

        for feature_score, weight in zip(feature_scores, self.weights):
            score += weight * feature_score

        return score



"""eFeature classes"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import logging


logger = logging.getLogger(__name__)


class EFeature(object):

    """EPhys feature"""

    def __init__(self, name):
        """Constructor"""

        self.name = name


class eFELFeature(EFeature):

    """eFEL feature"""

    def __init__(
            self,
            name,
            efel_feature_name=None,
            recording_names=None,
            stim_start=None,
            stim_end=None,
            exp_mean=None,
            exp_std=None,
            threshold=None):
        """Constructor"""

        EFeature.__init__(self, name)

        self.recording_names = recording_names
        self.efel_feature_name = efel_feature_name
        self.exp_mean = exp_mean
        self.exp_std = exp_std
        self.stim_start = stim_start
        self.stim_end = stim_end
        self.threshold = threshold

    def _construct_efel_trace(self, responses):
        """Construct trace that can be passed to eFEL"""

        trace = {}
        if '' not in self.recording_names:
            raise Exception(
                'eFELFeature: \'\' needs to be in recording_names')
        for location_name, recording_name in self.recording_names.iteritems():
            if location_name == '':
                postfix = ''
            else:
                postfix = ';%s' % location_name
            trace['T%s' % postfix] = \
                responses[self.recording_names['']]['time']
            trace['V%s' % postfix] = responses[recording_name]['voltage']
            trace['stim_start%s' % postfix] = [self.stim_start]
            trace['stim_end%s' % postfix] = [self.stim_end]

        return trace

    def calculate_feature(self, responses):
        """Calculate feature value"""

        efel_trace = self._construct_efel_trace(responses)

        import efel
        return efel.getMeanFeatureValues(
            [efel_trace],
            [self.efel_feature_name],
            raise_warnings=False)[0][self.efel_feature_name]

    def calculate_score(self, responses):
        """Calculate the score"""

        efel_trace = self._construct_efel_trace(responses)

        import efel

        if self.threshold:
            efel.setThreshold(self.threshold)

        score = efel.getDistance(
            efel_trace,
            self.efel_feature_name,
            self.exp_mean,
            self.exp_std)

        logger.debug('Calculated score for %s: %f', self.name, score)

        return score



"""Optimisation class"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0912


import random

import logging

import deap
import deap.creator
import deap.base
import deap.algorithms
import deap.tools
import pickle

logger = logging.getLogger('__main__')

# TODO decide which variables go in constructor, which ones go in 'run' function
# TODO abstract the algorithm by creating a class for every algorithm, that way
# settings of the algorithm can be stored in objects of these classes


class Optimisation(object):

    """Optimisation class"""

    def __init__(self, evaluator=None, eval_function=None,
                 use_scoop=False,
                 offspring_size=10):
        """Constructor"""
        self.evaluator = evaluator
        self.eval_function = eval_function
        self.model_params = evaluator.params
        self.objectives = evaluator.objectives
        self.use_scoop = use_scoop
        self.offspring_size = offspring_size
        # Create a DEAP toolbox
        self.toolbox = deap.base.Toolbox()

        self.setup_deap()

    def setup_deap(self):
        """Set up optimisation"""

        # Number of objectives
        OBJ_SIZE = len(self.objectives)

        class Individual(list):

            """Individual"""

            def __init__(self, *args):
                list.__init__(self, *args)
                self.time = None
                self.voltage = None

        class WeightedSumFitness(deap.base.Fitness):

            """Fitness that compares by weighted sum"""

            def __init__(self, values=()):
                deap.base.Fitness.__init__(self, values)

            @property
            def weighted_sum(self):
                """Weighted sum of wvalues"""
                return sum(self.wvalues)

            @property
            def sum(self):
                """Weighted sum of values"""
                return sum(self.values)

            def __le__(self, other):
                return self.weighted_sum <= other.weighted_sum

            def __lt__(self, other):
                return self.weighted_sum < other.weighted_sum

        # Create a fitness function
        # By default DEAP selector will try to maximise fitness values,
        # so we add a -1 weight value to minise
        deap.creator.create("Fitness",
                            WeightedSumFitness,
                            weights=[-1.0] * OBJ_SIZE)

        # Create an individual that consists of a list
        deap.creator.create(
            "Individual",
            Individual,
            fitness=deap.creator.Fitness)

        # Set random seed
        random.seed(1)

        # Eta parameter of crossover / mutation parameters
        # Basically defines how much they 'spread' solution around
        # The higher this value, the more spread
        ETA = 10.0

        # Number of parameters
        IND_SIZE = len(self.model_params)

        # Bounds for the parameters

        LOWER = []
        UPPER = []

        for parameter in self.model_params:
            LOWER.append(parameter.lower_bound)
            UPPER.append(parameter.upper_bound)

        # Define a function that will uniformly pick an individual
        def uniform(lower_list, upper_list, dimensions):
            """Fill array """

            if hasattr(lower_list, '__iter__'):
                return [random.uniform(lower, upper) for lower, upper in
                        zip(lower_list, upper_list)]
            else:
                return [random.uniform(lower_list, upper_list)
                        for _ in range(dimensions)]

        # Register the 'uniform' function
        self.toolbox.register("uniformparams", uniform, LOWER, UPPER, IND_SIZE)

        # Register the individual format
        # An indiviual is create by 'deap.creator.Individual' and parameters
        # are initially
        # picked by 'uniform'
        self.toolbox.register(
            "Individual",
            deap.tools.initIterate,
            deap.creator.Individual,
            self.toolbox.uniformparams)

        # Register the population format. It is a list of individuals
        self.toolbox.register(
            "population",
            deap.tools.initRepeat,
            list,
            self.toolbox.Individual)

        # Register the evaluation function for the individuals
        # import deap_efel_eval1
        self.toolbox.register("evaluate", self.eval_function)

        # Register the mate operator
        self.toolbox.register(
            "mate",
            deap.tools.cxSimulatedBinaryBounded,
            eta=ETA,
            low=LOWER,
            up=UPPER)

        # Register the mutation operator
        self.toolbox.register(
            "mutate",
            deap.tools.mutPolynomialBounded,
            eta=ETA,
            low=LOWER,
            up=UPPER,
            indpb=0.1)

        # Register the variate operator
        self.toolbox.register("variate", deap.algorithms.varAnd)

        # Register the selector (picks parents from population)
        self.toolbox.register("select", deap.tools.selIBEA)

        # toolbox.register("select", deap.tools.selIBEA)
        # toolbox.register("select", deap.tools.selIBEA)

        if self.use_scoop:
            from scoop import futures
            self.toolbox.register("map", futures.map)

            def _reduce_method(meth):
                """Overwrite reduce"""
                return (getattr, (meth.__self__, meth.__func__.__name__))
            import copy_reg
            import types
            copy_reg.pickle(types.MethodType, _reduce_method)

        # import multiprocessing

        # pool = multiprocessing.Pool()
        # self.toolbox.register("map", pool.map)

    def run(self, max_ngen=10, continue_cp=False, cp_filename=None):
        """Run optimisation"""

        # Total number of generation to run
        NGEN = max_ngen

        # Crossover, mutation probabilities
        CXPB = 0.7
        # MUTPB = 0.3

        # Total population size of EA
        # ALPHA = POP_SIZE
        # Total parent population size of EA
        MU = self.offspring_size
        # Total offspring size of EA
        # LAMBDA = OFFSPRING_SIZE

        # Generate the population object
        pop = self.toolbox.population(n=MU)

        hof = deap.tools.HallOfFame(10)

        stats = deap.tools.Statistics(key=lambda ind: ind.fitness.sum)

        import numpy
        stats.register("avg", numpy.mean)
        stats.register("std", numpy.std)
        stats.register("min", numpy.min)
        stats.register("max", numpy.max)

        pop, log, history = eaAlphaMuPlusLambdaCheckpoint(
            pop,
            self.toolbox,
            MU,
            None,
            CXPB,
            1 - CXPB,
            NGEN,
            stats=stats,
            halloffame=hof,
            continue_cp=continue_cp,
            cp_filename=cp_filename)

        return pop, hof, log, history

# pylint: disable=R0912, R0914

# TODO this function shouldn't be here


def eaAlphaMuPlusLambdaCheckpoint(
        population,
        toolbox,
        mu,
        _,
        cxpb,
        mutpb,
        ngen,
        stats=None,
        halloffame=None,
        checkpoint_freq=1,
        cp_filename=None,
        continue_cp=False):
    r"""This is the :math:`(~\alpha,\mu~,~\lambda)` evolutionary algorithm."""

    if continue_cp:
        # A file name has been given, then load the data from the file
        cp = pickle.load(open(cp_filename, "r"))
        population = cp["population"]
        parents = cp["parents"]
        start_gen = cp["generation"]
        halloffame = cp["halloffame"]
        logbook = cp["logbook"]
        history = cp["history"]
        random.setstate(cp["rndstate"])
    else:
        # Start a new evolution
        start_gen = 1
        parents = population[:]
        logbook = deap.tools.Logbook()
        logbook.header = ['gen', 'nevals'] + (stats.fields if stats else [])
        history = deap.tools.History()

        # TODO this first loop should be not be repeated !

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in population if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        record = stats.compile(population) if stats is not None else {}
        logbook.record(gen=start_gen, nevals=len(invalid_ind), **record)

        # Update the hall of fame with the generated individuals
        if halloffame is not None:
            halloffame.update(population)

        if history is not None:
            history.update(population)

    if history is not None:
        toolbox.decorate("mate", history.decorator)
        toolbox.decorate("mutate", history.decorator)

    # Begin the generational process
    for gen in range(start_gen + 1, ngen + 1):
        # Vary the parents
        if hasattr(toolbox, 'variate'):
            offspring = toolbox.variate(parents, toolbox, cxpb, mutpb)
        else:
            offspring = deap.algorithms.varAnd(parents, toolbox, cxpb, mutpb)

        population[:] = parents + offspring

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)

        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # Update the hall of fame with the generated individuals
        if halloffame is not None:
            halloffame.update(offspring)

        if history is not None:
            history.update(offspring)

        # Select the next generation parents
        parents[:] = toolbox.select(population, mu)

        # Update the statistics with the new population
        record = stats.compile(population) if stats is not None else {}
        logbook.record(gen=gen, nevals=len(invalid_ind), **record)

        print logbook.stream

        if cp_filename and checkpoint_freq:
            if gen % checkpoint_freq == 0:
                cp = dict(population=population, generation=gen,
                          parents=parents,
                          halloffame=halloffame,
                          history=history,
                          logbook=logbook, rndstate=random.getstate())
                pickle.dump(
                    cp,
                    open(cp_filename, "wb"))
                logger.debug('Wrote checkpoint to %s', cp_filename)

    return population, logbook, history



"""Run simple cell optimisation"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
# pylint: disable=R0914

import collections

import bluepyopt.electrical as nrpel
import bluepyopt.electrical.celltemplate as ct


# TODO store definition dicts in json
# TODO rename 'score' into 'objective'
# TODO add functionality to read settings of every object from config format

def define_mechanisms():
    """Define mechanisms"""

    import json
    with open('mechanisms.json') as mech_file:
        mech_definitions = json.load(mech_file)

    mechanisms = []
    for sectionlist, channels in mech_definitions.iteritems():
        seclist_loc = nrpel.locations.NrnSeclistLocation(
            sectionlist,
            seclist_name=sectionlist)
        for channel in channels:
            mechanisms.append(nrpel.mechanisms.NrnMODMechanism(
                name='%s.%s' % (channel, sectionlist),
                mod_path=None,
                prefix=channel,
                locations=[seclist_loc],
                preloaded=True))

    return mechanisms


def define_parameters():
    """Define parameters"""

    import json
    parameters = []

    # TODO put exp equation in file

    uniform_scaler = nrpel.parameterscalers.NrnSegmentLinearScaler()
    exponential_scaler = nrpel.parameterscalers.NrnSegmentSomaDistanceScaler(
        distribution='(-0.8696 + 2.087*math.exp(({distance})*0.0031))*{value}')

    # Fixed section parameters
    # TODO check the order of executions of all parameters

    with open('fixed_params.json', 'r') as fixed_params_file:
        fixed_params_definitions = json.load(
            fixed_params_file,
            object_pairs_hook=collections.OrderedDict)

    for sectionlist, params in fixed_params_definitions.iteritems():
        if sectionlist == 'global':
            for param_name, value in params:
                parameters.append(
                    nrpel.parameters.NrnGlobalParameter(
                        name=param_name,
                        param_name=param_name,
                        frozen=True,
                        value=value))
        else:
            seclist_loc = nrpel.locations.NrnSeclistLocation(
                sectionlist,
                seclist_name=sectionlist)

            for param_name, value, dist in params:
                parameters.append(nrpel.parameters.NrnSectionParameter(
                    name='%s.%s' % (param_name, sectionlist),
                    param_name=param_name,
                    value_scaler=uniform_scaler,
                    value=value,
                    frozen=True,
                    locations=[seclist_loc]))

    # Compact parameter description
    # Format ->
    # - Root dictionary: keys = section list name,
    #                    values = parameter description array
    # - Parameter description array: prefix, parameter name, minbound, maxbound

    with open('params.json', 'r') as parameter_file:
        parameter_definitions = json.load(
            parameter_file,
            object_pairs_hook=collections.OrderedDict)

    for sectionlist, params in parameter_definitions.iteritems():
        seclist_loc = nrpel.locations.NrnSeclistLocation(
            sectionlist,
            seclist_name=sectionlist)

        for prefix, param_name, min_bound, max_bound, dist in params:
            if dist == 'uniform':
                scaler = uniform_scaler
            elif dist == 'exp':
                scaler = exponential_scaler
            parameters.append(nrpel.parameters.NrnRangeParameter(
                name='%s_%s.%s' % (param_name, prefix, sectionlist),
                param_name='%s_%s' % (param_name, prefix),
                value_scaler=scaler,
                bounds=[min_bound, max_bound],
                locations=[seclist_loc]))

    return parameters


def define_morphology():
    """Define morphology"""

    return nrpel.morphologies.NrnFileMorphology('morphology/C060114A7.asc',
                                                do_replace_axon=True)


def create():
    """Create cell template"""

    cell = ct.CellTemplate(
        'l5pc',
        morph=define_morphology(),
        mechs=define_mechanisms(),
        params=define_parameters())

    return cell



"""Stimuli classes"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import logging

import bluepyopt as nrp

logger = logging.getLogger(__name__)


class Stimulus(object):

    """Stimulus protocol"""

    def __init__(self):
        """Constructor"""
        pass

# TODO Add 'current' to the name


class NrnSquarePulse(Stimulus):

    """Stimulus protocol"""

    def __init__(self,
                 step_amplitude=None,
                 step_delay=None,
                 step_duration=None,
                 total_duration=None,
                 location=None):
        """Constructor"""

        Stimulus.__init__(self)
        self.step_amplitude = step_amplitude
        self.step_delay = step_delay
        self.step_duration = step_duration
        self.location = location
        self.total_duration = total_duration
        self.iclamp = None

    def instantiate(self, cell):
        """Run stimulus"""

        icomp = self.location.instantiate(cell)
        logger.debug(
            'Adding square step stimulus to %s with delay %f, '
            'duration %f, and amplitude %f',
            str(self.location),
            self.step_delay,
            self.step_duration,
            self.step_amplitude)

        self.iclamp = nrp.neuron.h.IClamp(
            icomp.x,
            sec=icomp.sec)
        self.iclamp.dur = self.step_duration
        self.iclamp.amp = self.step_amplitude
        self.iclamp.delay = self.step_delay

    def destroy(self):
        """Destroy stimulus"""

        self.iclamp = None

    def __str__(self):
        """String representation"""

        return "Square pulse amp %f delay %f duration %f totdur %f at %s" % (
            self.
            step_amplitude,
            self.
            step_delay,
            self.
            step_duration,
            self.
            total_duration,
            self.
            location)



"""Cell template class"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# pylint: disable=W0511

# TODO take into account that one might want to run protocols on different
# machines
# TODO rename this to 'CellModel' ?

import collections
import logging

import bluepyopt as bpopt

logger = logging.getLogger(__name__)


class CellTemplate(object):

    """Simple cell class"""

    def __init__(
            self,
            name,
            morph=None,
            mechs=None,
            params=None):
        """Constructor"""

        self.name = name

        # morphology
        self.morphology = morph
        # mechanisms
        self.mechanisms = mechs
        # Model params
        self.params = collections.OrderedDict()
        for param in params:
            self.params[param.name] = param

        # Cell instantiation in simulator
        self.icell = None

        self.param_values = None

    def freeze(self, param_values):
        """Set params"""

        for param_name, param_value in param_values.items():
            param = self.params[param_name]
            param.freeze(param_value)

    def unfreeze(self, param_names):
        """Unset params"""

        for param_name in param_names:
            self.params[param_name].unfreeze()

    @staticmethod
    def create_empty_cell(name):
        """Create an empty cell in Neuron"""

        # TODO minize hardcoded definition
        # E.g. sectionlist can be procedurally generated
        template_content = 'begintemplate %s\n' \
            'objref all, apical, basal, somatic, axonal\n' \
            'proc init() {\n' \
            'all 	= new SectionList()\n' \
            'somatic = new SectionList()\n' \
            'basal 	= new SectionList()\n' \
            'apical 	= new SectionList()\n' \
            'axonal 	= new SectionList()\n' \
            'forall delete_section()\n' \
            '}\n' \
            'create soma[1], dend[1], apic[1], axon[1]\n' \
            'endtemplate %s\n' % (name, name)

        bpopt.neuron.h(template_content)

        template_function = getattr(bpopt.neuron.h, name)

        return template_function()

    def instantiate(self):
        """Instantiate model in simulator"""

        bpopt.neuron.h.load_file('stdrun.hoc')

        # TODO replace this with the real template name
        if not hasattr(bpopt.neuron.h, 'Cell'):
            self.icell = self.create_empty_cell('Cell')
        else:
            self.icell = bpopt.neuron.h.Cell()

        self.morphology.instantiate(self)

        for mechanism in self.mechanisms:
            mechanism.instantiate(self)
        for param in self.params.values():
            param.instantiate(self)

    # TODO This should also get param_values as argument, and original
    # function renamed
    def run_protocol(self, protocol):
        """Run protocol"""

        self.instantiate()
        protocol.instantiate(self)

        bpopt.neuron.h.tstop = protocol.total_duration
        bpopt.neuron.h.cvode_active(1)
        logger.debug(
            'Running protocol %s for %.6g ms',
            protocol.name,
            protocol.total_duration)
        bpopt.neuron.h.run()
        responses = protocol.responses

        protocol.destroy()
        self.destroy()

        logger.debug('Protocol finished, returning responses')
        return responses

    def destroy(self):
        """Destroy instantiated model in simulator"""

        self.icell = None
        self.morphology.destroy()
        for mechanism in self.mechanisms:
            mechanism.destroy()
        for param in self.params.values():
            param.destroy()

    def check_nonfrozen_params(self, param_names):
        """Check if all nonfrozen params are set"""

        for param_name, param in self.params.items():
            if not param.frozen:
                raise Exception(
                    'CellTemplate: Nonfrozen param %s needs to be '
                    'set before simulation' %
                    param_name)

    def run_protocols(self, protocols, param_values=None):
        """Run stimulus protocols"""

        # TODO Put all of this in a decorator ?
        import traceback
        import sys
        try:
            self.freeze(param_values)

            self.check_nonfrozen_params(param_values.keys())

            responses = {}
            for protocol in protocols.itervalues():
                protocol_responses = self.run_protocol(protocol)
                for response_name, response in protocol_responses.iteritems():
                    if response_name in responses:
                        raise Exception(
                            'CellTemplate: response name used twice: %s' %
                            response.name)
                    responses[response_name] = response

            self.unfreeze(param_values.keys())
        except:
            raise Exception(
                "".join(
                    traceback.format_exception(
                        *
                        sys.exc_info())))

        return responses

    def __str__(self):
        """Return string representation"""

        content = '%s:\n' % self.name

        content += '  morphology:\n'
        content += '    %s\n' % str(self.morphology)

        content += '  mechanisms:\n'
        for mechanism in self.mechanisms:
            content += '    %s\n' % mechanism
        content += '  params:\n'
        for param in self.params.values():
            content += '    %s\n' % param

        return content



"""Protocol classes"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# TODO: maybe find a better name ? -> sweep ?


class Protocol(object):

    """Stimulus protocol"""

    def __init__(self, name=None, stimuli=None, recordings=None):
        """Constructor"""
        self.name = name
        self.stimuli = stimuli
        self.recordings = recordings

    @property
    def total_duration(self):
        """Total duration"""

        return max([stimulus.total_duration for stimulus in self.stimuli])

    @property
    def responses(self):
        """Return all the responses"""
        return {
            recording.name: recording.response for recording in self.recordings}

    def instantiate(self, cell):
        """Instantiate protocol"""

        for stimulus in self.stimuli:
            stimulus.instantiate(cell)

        for recording in self.recordings:
            recording.instantiate(cell)

    def destroy(self):
        """Destroy protocol"""

        for stimulus in self.stimuli:
            stimulus.destroy()

        for recording in self.recordings:
            recording.destroy()

    def __str__(self):
        """String representation"""

        content = '%s:\n' % self.name

        content += '  stimuli:\n'
        for stimulus in self.stimuli:
            content += '    %s\n' % str(stimulus)

        content += '  recordings:\n'
        for recording in self.recordings:
            content += '    %s\n' % str(recording)

        return content



"""Run simple cell optimisation"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

"""
This optimisation is based on L5PC optimisations developed by Etay Hay in the
context of the BlueBrain project
"""


# pylint: disable=R0914

import bluepyopt

# TODO store definition dicts in json
# TODO rename 'score' into 'objective'
# TODO add functionality to read settings of every object from config format

import l5pc_evaluator
evaluator = l5pc_evaluator.create()


def evaluate(parameter_array):
    """Global evaluate function"""

    return evaluator.evaluate(parameter_array)

opt = bluepyopt.Optimisation(
    evaluator=evaluator,
    eval_function=evaluate,
    offspring_size=2,
    use_scoop=True)


def main():
    """Main"""

    import argparse
    parser = argparse.ArgumentParser(description='L5PC example')
    parser.add_argument('--start', action="store_true")
    parser.add_argument('--continue_cp', action="store_true")
    parser.add_argument('--analyse', action="store_true")

    # TODO read checkpoint filename from arguments
    cp_filename = 'checkpoint.pkl'

    args = parser.parse_args()

    if args.start or args.continue_cp:
        opt.run(
            max_ngen=200,
            continue_cp=args.continue_cp,
            cp_filename=cp_filename)

    if args.analyse:
        import l5pc_analysis
        l5pc_analysis.analyse_cp(opt=opt, cp_filename=cp_filename)
        l5pc_analysis.analyse_releasecircuit_model(opt=opt)

        import matplotlib.pyplot as plt
        plt.show()

if __name__ == '__main__':
    main()



"""
Mechanism classes

Theses classes represent mechanisms in the model
"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import logging

logger = logging.getLogger(__name__)

# TODO: use Location class to specify location


class Mechanism(object):

    """Base parameter class"""

    def __init__(self, name):
        """Constructor"""

        self.name = name


class NrnMODMechanism(Mechanism):

    """Neuron mechanism"""

    def __init__(
            self,
            name,
            mod_path=None,
            prefix=None,
            locations=None,
            preloaded=True):
        """Constructor"""

        Mechanism.__init__(self, name)
        self.mod_path = mod_path
        self.prefix = prefix
        self.locations = locations
        self.preloaded = True

    def instantiate(self, cell):
        """Instantiate"""

        for location in self.locations:
            isec_list = location.instantiate(cell)
            for isec in isec_list:
                isec.insert(self.prefix)
            logger.debug('Inserted %s in %s',
                         self.prefix, self.locations)

    def destroy(self):
        """Destroy mechanism instantiation"""

        pass

    def __str__(self):
        """String representation"""

        return "%s: %s %s" % (self.name, self.locations, self.prefix)



"""Cell evaluator class"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# pylint: disable=W0511

# Not sure code below should go in module or class
import copy_reg
import types


class CellEvaluator(object):

    """Simple cell class"""

    def __init__(
            self,
            cell_template=None,
            param_names=None,
            fitness_protocols=None,
            fitness_calculator=None,
            isolate_protocols=True):
        """Constructor"""

        self.cell_template = cell_template
        self.param_names = param_names
        # Stimuli used for fitness calculation
        self.fitness_protocols = fitness_protocols
        # Fitness value calculator
        self.fitness_calculator = fitness_calculator

        self.isolate_protocols = isolate_protocols

    @property
    def objectives(self):
        """Return objectives"""

        return self.fitness_calculator.objectives

    @property
    def params(self):
        """Return params of this evaluation"""

        params = []
        for param_name in self.param_names:
            params.append(self.cell_template.params[param_name])

        return params

    def param_dict(self, param_array):
        """Convert param_array in param_dict"""
        param_dict = {}
        for param_name, param_value in \
                zip(self.param_names, param_array):
            param_dict[param_name] = param_value

        return param_dict

    def objective_dict(self, objective_array):
        """Convert objective_array in objective_dict"""
        objective_dict = {}
        objective_names = [objective.name
                           for objective in self.fitness_calculator.objectives]
        for objective_name, objective_value in \
                zip(objective_names, objective_array):
            objective_dict[objective_name] = objective_value

        return objective_dict

    def evaluate_with_dicts(self, param_dict=None):
        """Run evaluation with dict as input and output"""

        if self.fitness_calculator is None:
            raise Exception('CellTemplate: need fitness_calculator to evaluate')

        responses = {}

        # TODO clean this up

        if self.isolate_protocols:
            import multiprocessing

            # TODO this should only be executed once

            def _reduce_method(meth):
                """Overwrite reduce"""
                return (getattr, (meth.__self__, meth.__func__.__name__))

            copy_reg.pickle(types.MethodType, _reduce_method)

        for protocol_name, protocol in self.fitness_protocols.iteritems():
            if self.isolate_protocols:
                # This multiprocessing makes sure that Neuron starts in a clean
                # state every time we run a protocol
                pool = multiprocessing.Pool(1, maxtasksperchild=1)
                responses.update(
                    pool.apply(self.cell_template.run_protocols,
                               args=[{protocol_name: protocol}],
                               kwds={'param_values': param_dict}))

                # THis might help with garbage collecting the pool workers
                pool.terminate()
                pool.join()
            else:
                responses.update(self.cell_template.run_protocols(
                    {protocol_name: protocol},
                    param_values=param_dict))

        return self.fitness_calculator.calculate_scores(responses)

    def evaluate_with_lists(self, param_list=None):
        """Run evaluation with lists as input and outputs"""

        param_dict = self.param_dict(param_list)

        obj_dict = self.evaluate_with_dicts(param_dict=param_dict)

        return obj_dict.values()
